open Exceptions
open Functions.Impl

module type Intf = sig
  (** Applies all pending migrations to the database in the order
      they've been given.

      Stores the code of all applied migrations in [_migrations_] table.
      Does not apply a migration if it has already been applied and its
      current code matches the one stored in the database.

      Raises an exception whenever SQLite does not return OK.
      Raises an exception when the code of an already applied migration
      under the index [i] differs from the code in [migrations(i)].
      Re-raises all other underlying exceptions from Sqlite3.
  *)
  val apply: Sqlite3.db -> migrations: string list -> unit

  (** Checks that all migrations are applied to the database. *)
  val check: Sqlite3.db -> migrations: string list -> bool
end

module Impl: Intf = struct
  open Core

  let create_migrations_table db =
    exec_only db {|
    CREATE TABLE IF NOT EXISTS _migrations_
    ( id integer PRIMARY KEY
    , code text NOT NULL
    , created_at text NOT NULL);|}

  let fetch_migration db id =
    let open Sqlite3.Data in
    let rows = exec_fetch (module From_sql.Id) db
        "SELECT code FROM _migrations_ WHERE id = :id;"
        ~params:[":id", INT (Int64.of_int id)]
    in
    match[@warning "-4"] rows with
    | [|[| "code", TEXT code |]|] -> Some code
    | [| |] -> None
    | _ -> raise (Corrupted_migration_entry id)

  let fail_on_code_mismatch id code_in_db current_code = match code_in_db with
    | Some code when not (String.equal code current_code) ->
      raise (Codes_differ (id, code, current_code))
    | _ -> ()

  let store_migration db id code =
    let open Sqlite3.Data in
    exec_only db
      {|INSERT OR REPLACE
            INTO _migrations_ (id, code, created_at)
            VALUES (:id, :code, :created_at);|}
      ~params:[
        ":id", INT (Int64.of_int id);
        ":code", TEXT code;
        ":created_at", TEXT (Time.to_string (Time.now ()));
      ]

  let apply db ~migrations:migrations =
    create_migrations_table db;
    List.iteri migrations ~f:(fun id current_code ->
        let code_in_db = fetch_migration db id in
        fail_on_code_mismatch id code_in_db current_code;
        if Option.is_none code_in_db
        then begin
          exec_only db current_code;
          store_migration db id current_code
        end
      )

  let migrations_table_exists db =
    try
      exec_only db "SELECT id, code, created_at from _migrations_;";
      true
    with
    | Sqlite_error _ -> false

  let all_migrations_are_applied db migrations =
    try
      List.for_alli migrations ~f:begin fun id current_code ->
        let code_in_db = fetch_migration db id in
        fail_on_code_mismatch id code_in_db current_code;
        Option.is_some code_in_db
      end
    with
    | Sqlite_error _ | Codes_differ _ -> false

  let check db ~migrations:migrations =
    migrations_table_exists db && all_migrations_are_applied db migrations
end
