open Core
open Oksql
open Alcotest

let with_db_in_mem fn =
  with_db ":memory:" fn

let has_table db table_name =
  let results = exec_fetch (module From_sql.Int64) db
      "SELECT count(*) = 1 FROM sqlite_master WHERE type = 'table' AND name = :name;"
      ~params:[":name", Sqlite3.Data.TEXT table_name] in
  Int64.equal results.(0) 1L

let test_should_create_migrations_table =
  test_case "should create _migrations_ table if it doesn't exist" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        Migration.apply db ~migrations:[];
        (check bool) "has table '_migrations_'" true (has_table db "_migrations_")
      )
    end

let test_should_apply_migrations_in_order =
  test_case "should apply migrations in the given order" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        Migration.apply db ~migrations:[
          "CREATE TABLE foo (bar integer);";
          "ALTER TABLE foo RENAME TO bar;";
        ];
        (check bool) "has table 'bar'" true (has_table db "bar");
        (check bool) "doesn't have table 'foo'" false (has_table db "foo");
      )
    end

let test_should_fail_if_migration_codes_differ =
  test_case "should fail if migration codes differ" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        Migration.apply db ~migrations:[];
        let result = Sqlite3.exec db "INSERT INTO _migrations_ VALUES (0, 'foo', 0);" in
        (check bool) "migration was inserted" true (phys_equal Sqlite3.Rc.OK result);
        (check_raises "migration codes differ"
           (Codes_differ (0, "foo", "bar"))
           (fun () -> Migration.apply db ~migrations:["bar"]))
      )
    end

let test_should_fail_when_sqlite_fails =
  test_case "should fail when sqlite fails" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        check_raises "sqlite error"
          (Sqlite_error (Sqlite3.Rc.ERROR, "preparing 'some garbage': Sqlite3.prepare: near \"some\": syntax error"))
          (fun () -> Migration.apply db ~migrations:["some garbage"])
      )
    end

let test_should_fail_when_migrations_table_is_corrupted =
  test_case "should fail when _migrations_ table is corrupted" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        ignore (Sqlite3.exec db "CREATE TABLE _migrations_ (id integer, code integer, created_at integer);");
        ignore (Sqlite3.exec db "INSERT INTO _migrations_ VALUES (0, 0, 0);");
        check_raises "migrations table corrupted"
          (Corrupted_migration_entry 0)
          (fun () -> Migration.apply db ~migrations:["foo"])
      )
    end

let test_should_set_created_at_when_applying_migration =
  test_case "should set 'created_at' to the current timestamp when applying migration" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        let ts_before_test = Time.now () in
        Migration.apply db ~migrations:["CREATE TABLE foo (bar integer);"];
        let results = exec_fetch (module From_sql.String) db "SELECT created_at FROM _migrations_ LIMIT 1;" in
        let ts_in_db = Time.of_string results.(0) in
        let ts_after_test = Time.now () in
        (check bool) "timestamp in db is later than before test" true
          (Time.is_later ts_in_db ~than:ts_before_test);
        (check bool) "timestamp in db is earlier than after test" true
          (Time.is_earlier ts_in_db ~than:ts_after_test);
      )
    end

let test_should_not_apply_already_applied_migrations =
  test_case "should not apply already applied migrations" `Quick
    begin fun () -> with_db_in_mem (fun db ->
        let migrations = [
          "CREATE TABLE foo (bar integer PRIMARY KEY);";
          "INSERT INTO foo VALUES (1);";
        ] in
        Migration.apply db ~migrations:migrations;
        (* No exception should be raised *)
        Migration.apply db ~migrations:(
          List.append migrations ["INSERT INTO foo VALUES (2);"])
      )
    end

let test_should_check_migrations =
  test_case "should check if migrations are applied to the db" `Quick
    begin fun () -> with_db_in_mem begin fun db ->
        let migrations = [ "CREATE TABLE foo (foo integer);";
                           "CREATE TABLE bar (bar text);"; ] in
        (check bool) "empty database is checked as ok" false
          (Migration.check db ~migrations:migrations);
        Migration.apply db ~migrations:(List.take migrations 1);
        (check bool) "database with partially applied migrations is checked as ok" false
          (Migration.check db ~migrations:migrations);
        Migration.apply db ~migrations:migrations;
        (check bool) "database with fully applied migrations is checked as not ok" true
          (Migration.check db ~migrations:migrations);
      end
    end

let cases = [
  test_should_create_migrations_table;
  test_should_apply_migrations_in_order;
  test_should_fail_if_migration_codes_differ;
  test_should_fail_when_sqlite_fails;
  test_should_fail_when_migrations_table_is_corrupted;
  test_should_set_created_at_when_applying_migration;
  test_should_not_apply_already_applied_migrations;
  test_should_check_migrations;
]
